class JoinError(Exception):
    """Represents an error when joining two primitives."""

    pass


class MeetError(Exception):
    """Represents an error when finding the intersection of two primitives."""

    pass
